#ifndef ALIGNER_HH
#define ALIGNER_HH

#include "sequence.hh"
#include "basepairs.hh"
#include "alignment.hh"

#include "params.hh"
#include "scoring.hh"

#include "matrices.hh"



/*
  TODO: score matrices should be as small as possible
  and have offsets.
  The M-Matrix can be smaller if the maximal arc len is
  limited. The D-matrix may be smaller if the number
  of simultaneously needed arc-pairs is limited, due
  to limit on the local sub-sequence lengths.
  
  A special cyclically rotatable matrix is needed for
  alignment on the top-level.
*/


/**
   Contains data for restricting Aligner to sub-sequences startA..endA amd startB..endB.

   Take care when using aligner restrictions for multiple
   Alignments with the same aligner object.
   The D-matrix is only computed once, so this works
   as long as the first Aligner::align() is called with
   the most general restriction (e.g. no restriction at all)!
*/
class AlignerRestriction {
private:
    int startA;
    int startB;
    int endA;
    int endB;
public:
    AlignerRestriction(int startA_, int startB_, int endA_, int endB_)
	: startA(startA_), startB(startB_), endA(endA_), endB(endB_)
    {}

    size_t get_startA() const {return startA;}
    size_t get_endA() const {return endA;}
    size_t get_startB() const {return startB;}
    size_t get_endB() const {return endB;}

    void set_startA(size_t p) {startA=p;}
    void set_endA(size_t p)  {endA=p;}
    void set_startB(size_t p)  {startB=p;}
    void set_endB(size_t p)  {endB=p;}
};

inline
std::ostream & operator<<(std::ostream &out, AlignerRestriction r) {
    return
	out << r.get_startA() << " "
	    << r.get_startB() << " "
	    << r.get_endA() << " "
	    << r.get_endB();
}


/* ============================================================ 
   class Aligner

   Compute alignment similarity score and optimal alignment
   
   initialize with two sequences and their basepairs (including scores)
   ============================================================ 
*/

#define FOR_IJ_IN_ARCMATCH \
  for (int i=al+1; i<ar; i++) \
    for (int j=bl+1; j<br; j++)

//! class for doing the alignment of two sequences and their
//! associated sets of weighted basepairs
//!
//! an object always knows about the two sequences and
//! the two weighted base pair sets
//!
//! common usage: construct, align, trace, get_alignment
class Aligner {
public:
    typedef size_t size_type;
    typedef ScoreMatrix M_matrix_t;
    // typedef RMtrix<infty_score_t> M_matrix_t; // didn't improve performance

protected:
    const Scoring *scoring; //!< the scores
    
    const AlignerParams *params; //!< the parameter for the alignment
	
    Sequence &seqA;
    BasePairs &bpsA;
    Sequence &seqB;
    BasePairs &bpsB;

    // The following is added for the idea of
    // getting the k-best local alignments
    //
    // Aligner always works on sub-sequences/structures of
    // seqA and seqB. These are specified by
    AlignerRestriction r;
    // in the standard case, i.e. align the whole sequences,
    // the values are set to startA=1 and endA=seqA.length()
    // and analogously for B
	
    // the constructor initializes to the above standard values
    // methods provide possibility to restriction
  
    ScoreMatrix D;
    //DMatrix<score_t> D;
    std::vector<M_matrix_t> Ms;
    
    //! for cool affine gap cost,
    //! we need two additional matrices E and F
    //! however we only need to store one row or column.
    //!
    //! for structure local, we need one such matrix per state
    std::vector<ScoreVector> Es;
    std::vector<ScoreVector> Fs;

    
    int max_i,max_j; // computed by align_top_level
    
    bool D_created; // flag, is D already created?

    Alignment alignment;




    enum {E_NO_NO, E_X_NO, E_NO_X, E_X_X,
	  E_OP_NO, E_NO_OP, E_OP_X, E_X_OP};
  

    //! initialize first column and row
    void init_state(int state, int al, int ar, int bl, int br, 
		    bool globalA, bool exclA,
		    bool globalB, bool exclB);
    
    // recursion cases that handle everything but exclusions
    // (in the LSSA-paper this function was called NoEx
    
    //! standard cases for alignment (without exlusion handling)
    infty_score_t align_noex(int state, size_type al, size_type bl, size_type i, size_type j);
     
    //! align, where al,ar,bl,br is an arc-match
    //! in structure local alignment, this allows to introduce exclusions
    void align_in_arcmatch(int al,int ar,int bl,int br,
			   bool allow_exclusion);
  
    //! align the top-level in a sequence local alignment
    //! and return the maximal score
    infty_score_t align_top_level_locally();

    //! align top level in the scanning version
    infty_score_t align_top_level_localB();
  
    //! trace back within an match of arcs
    void trace_in_arcmatch(int state,int al,int i,int bl,int j,bool top_level);
  
    //! standard cases in trace back (without handling of exclusions)
    void trace_noex(int state,size_type al,size_type i,size_type bl,size_type j,bool top_level);

    //! trace an arc match
    void trace_arcmatch(const Arc &arcA, const Arc &arcB);

    //! trace an arc match in case of forbidden lonely pairs
    void trace_arcmatch_noLP(const Arc &arcA, const Arc &arcB);

    /**
       create the entries in the D matrix
       This function is called by align() (unless D_created)
    */
    void align_D();

    /**
       fill in D the entries with left ends al,bl, 
       where adjlA, adjlB are adjanceny lists of the arcs
    */
    void 
    fill_D_entries(size_type al, size_type bl,
		   const BasePairs::LeftAdjList &adjlA,
		   const BasePairs::LeftAdjList &adjlB,
		   size_type max_ar, size_type max_br
		   );
    /**
       fill D entries when no-lonely-pairs option given
       after computation of M-matrices with left ends al,bl
       
       this fills D entries for arc matches with left ends al-1,bl-1,
       since the positions refer to the inner arc
       of a stacked arc pair
    */
    void 
    fill_D_entries_noLP(size_type al, size_type bl, size_type max_ar, size_type max_br);

public:  
    //! copy constructor
    Aligner(const Aligner &a);

    //! construct with sequences and corresponding sets of basepairs
    Aligner(Sequence &seqA, BasePairs &bpsA,
	    Sequence &seqB, BasePairs &bpsB,
	    const AlignerParams *ap,
	    const Scoring *s
	    );
    
    //! return the alignment that was computed by trace()
    Alignment const & get_alignment() const {return alignment;} 
    
    //! assignment operator
    const Aligner & operator = (const Aligner &a);
    
    //! compute the alignment score
    infty_score_t align();
    
    //! do the trace back through the alignment matrix
    //! with partial recomputation
    //! pre: call align() to fill the top-level matrix
    void trace();
    
    //! set the restriction on the alignment,
    //! mainly used for the k-best algorithm
    void set_restriction(const AlignerRestriction &r_) {r=r_;}
    
    //! return the current restriction,
    //! mainly used for the k-best algorithm
    AlignerRestriction &get_restriction() {return r;}

  
    void kbest(size_t k,
	       size_t output_width,
	       bool opt_verbose,
	       bool opt_local_out,
	       bool opt_pos_output
	       );
};

#endif
