#include "aligner_p.hh"

// #include "d_matrix.hh"

#include <math.h>
#include <assert.h>
#include<iomanip>
#include <queue>

using namespace std;

// ------------------------------------------------------------
// AlignerP: compute partition function and probabilities of arc matchs and base matchs
//

/*
//===========================================================================
//	alignerP()
//===========================================================================	
AlignerP::AlignerP(const AlignerP &a) :
	scoring(a.scoring), params(a.params), seqA(a.seqA), bpsA(a.bpsA),
	seqB(a.seqB), bpsB(a.bpsB), r(a.r), D(a.D), M(a.M),
	MPrime(a.MPrime), DPrime(a.DPrime), Mrev(a.Mrev),am_prob(a.am_prob),bm_prob(a.bm_prob),
	max_i(a.max_i), max_j(a.max_j), D_created(a.D_created),
	DPrime_created(a.DPrime_created),
	alignment(a.alignment) {
}

//===========================================================================
//	alignerP()
//===========================================================================	
const AlignerP & AlignerP::operator = (const AlignerP &a) {
	params=a.params;
	r=a.r;
	D=a.D;
	M=a.M;
	DPrime=a.DPrime;
	MPrime=a.MPrime;
	Mrev=a.Mrev;
	am_prob=a.am_prob;
	bm_prob=a.bm_prob;
	seqA=a.seqA;
	bpsA=a.bpsA;
	seqB=a.seqB;
	bpsB=a.bpsB;
	alignment=a.alignment;
	max_i=a.max_i;
	max_j=a.max_j;
	D_created=a.D_created;
	DPrime_created=a.DPrime_created;
	scoring=a.scoring;
	return a;
}
*/

//===========================================================================
//	alignerP()
//===========================================================================	
AlignerP::AlignerP(Sequence &seqA_, BasePairs &bpsA_, Sequence &seqB_,
		BasePairs &bpsB_, const AlignerParams *ap_, const Scoring *s_) :
			scoring(s_), params(ap_), seqA(seqA_), bpsA(bpsA_), seqB(seqB_),
			bpsB(bpsB_), r(1, 1, seqA_.length(), seqB_.length()),
			D_created(false), DPrime_created(false),
			alignment(seqA_, seqB_) {

	D.resize(bpsA.num_bps(), bpsB.num_bps());
	D.fill((pf_score_t )0);
	M.resize(seqA.length()+1, seqB.length()+1);
	M.fill((pf_score_t )0);

	DPrime.resize(bpsA.num_bps(), bpsB.num_bps());
	DPrime.fill((pf_score_t )0);

	MPrime.resize(seqA.length()+1, seqB.length()+1);
	MPrime.fill((pf_score_t )0);

	Mrev.resize(seqA.length()+2, seqB.length()+2); // "+2" necessary for computing reverse of whole alig
	Mrev.fill((pf_score_t )0);

	am_prob.resize(bpsA.num_bps(), bpsB.num_bps());
	am_prob.fill(0.0);
	bm_prob.resize(seqA.length()+1, seqB.length()+1);
	bm_prob.fill(0.0);

	am_heuristic = (params->max_diff>=0) || (params->max_diff_am>=0); 
}

//===========================================================================
//	align_noex_inside()
//===========================================================================	
//! standard cases in alignment: base match, base in/del, arc match
inline
pf_score_t
AlignerP::align_noex_inside(size_type al, size_type bl, size_type i, size_type j) {

	pf_score_t max_score;
	// base match
	max_score = M(i-1, j-1) * scoring->exp_basematch(i, j);

	// base del
	max_score += M(i-1, j) * scoring->exp_gapA(i, j);

	// base ins
	max_score += M(i, j-1) * scoring->exp_gapB(i, j);

	// arc match

	// standard case for arc match (without restriction to lonely pairs)
	assert(max_score > 0);
	const BasePairs::RightAdjList &adjlA = bpsA.right_adjlist(i);
	const BasePairs::RightAdjList &adjlB = bpsB.right_adjlist(j);

	// for all pairs of arcs in A and B that have right ends i and j, respectively
	//
	for (BasePairs::RightAdjList::const_iterator arcA=adjlA.begin(); 
	     arcA !=adjlA.end() && arcA->left() > al; ++arcA) {
		for (BasePairs::RightAdjList::const_iterator arcB=adjlB.begin(); 
		     arcB !=adjlB.end() && arcB->left() > bl; ++arcB) {

			// consider score for match of basepairs
			assert(M(arcA->left()-1, arcB->left()-1) > 0);
			assert(am_heuristic || D(arcA->idx(), arcB->idx())>0);
			max_score += M(arcA->left()-1, arcB->left()-1) * D(arcA->idx(), arcB->idx());
			// note: disallowed arc matchs (due to heuristic) are
			// handled correctly, since there D(arcA->idx(), arcB->idx()) was set to 0
		}
	}

	return max_score;
}

//===========================================================================
//	init_M_inside()
//===========================================================================	
//! generic initalization method. The "0th" row and column are initialized
//
void AlignerP::init_M_inside(size_type al, size_type ar, size_type bl, size_type br) {

	M(al, bl)=(pf_score_t)1;

	pf_score_t indel_score=(pf_score_t)1;

	for (size_type i=al+1; i<ar; i++) {
		indel_score *= scoring->exp_gapA(i, bl);
		M(i, bl) = (pf_score_t)indel_score;
		assert(M(i,bl)>0);
	}

	indel_score=(pf_score_t)1;
	for (size_type j=bl+1; j<br; j++) {
		indel_score *= scoring->exp_gapB(al, j);
		M(al, j) = (pf_score_t)indel_score;
	}
}

//===========================================================================
//	align_in_arcmatch
//===========================================================================	
void AlignerP::align_in_arcmatch(size_type al, size_type ar, size_type bl, size_type br) {

	init_M_inside(al, ar, bl, br);

	for (size_type i=al+1; i<ar; i++)
		for (size_type j=bl+1; j<br; j++)
			M(i, j)=align_noex_inside(al,bl,i,j);
}

//===========================================================================
//	fill_D_entries_inside
//===========================================================================	

// compute the entries in the D matrix that
// can be computed from the matrix/matrices M
// for the subproblem al,bl,max_ar,max_br
//
// pre: M matrices are computed by a call to 
//      align_in_arcmatch(al,max_ar,bl,max_br)
//
void AlignerP::fill_D_entries_inside(size_type al, size_type bl,
		const BasePairs::LeftAdjList &adjlA,
		const BasePairs::LeftAdjList &adjlB, size_type max_ar, size_type max_br) {

	for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin(); 
	     arcA !=adjlA.end(); ++arcA) {

		for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin(); 
		     arcB !=adjlB.end(); ++arcB) {

			size_type ar = arcA->right();
			size_type br = arcB->right();
			//
			// if right ends ar,br exceed the limits max_ar,max_br resp.
			// the corresponding D entry is set to 0
			// in order to dissalow the arc match completely.
			// This occurs only due to an am heuristic.
			if (ar>max_ar || br>max_br) {
			    
			    assert(am_heuristic);
			    
			    D(arcA->idx(), arcB->idx()) = (pf_score_t)0;
			} else {
			    assert( M(ar-1, br-1) > 0 );
			    D(arcA->idx(), arcB->idx()) = M(ar-1, br-1) * scoring->exp_arcmatch(*arcA,*arcB);
			}
		}
	}
}

//===========================================================================
//	align_D_inside()
//===========================================================================	
// compute all entries D
void AlignerP::align_D_inside() {
	// ------------------------------------------------------------
	// General workflow:
	//
	// for all combinations of left arc ends al and ar
	// 1.) determie for which arc-pairs the D entries can be computed   
	// in one run, 2.) call align_in_arcmatch 3.) call fill_D_entries
	// ------------------------------------------------------------

	// ------------------------------------------------------------
	// traverse the left ends al,bl of arcs in descending order
	//
    for (size_type al=r.get_endA(); al>=r.get_startA(); al--) {

	const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(al);
	
	if (adjlA.size()>0) {
	    
	    size_type min_bl=r.get_startB();
	    size_type max_bl=r.get_endB();

	    // if max-diff heuristic is active,
	    // restrict range for left ends of bl
	    if (params->max_diff>=0) {
		min_bl = std::max(min_bl+params->max_diff, al) - params->max_diff;
		max_bl = std::min(max_bl, al + params->max_diff);
	    }

	    for (size_type bl=max_bl; bl>=min_bl; bl--) {
		
		const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(bl);
		
		if (adjlB.size()>0) {
		    
		    // ------------------------------------------------------------
		    // get maximal right ends of arcs with left ends al,bl 
		    // where (if active) am-heuristic conditions hold
		    //
		    size_type max_ar=al;
		    size_type max_br=bl;
		    if (! am_heuristic ) {
			max_ar=adjlA[adjlA.size()-1].right();
			max_br=adjlB[adjlB.size()-1].right();
		    } else {
			for (BasePairs::LeftAdjList::const_iterator arcA =
				 adjlA.begin(); arcA!=adjlA.end(); ++arcA) {
			    for (BasePairs::LeftAdjList::const_iterator arcB =
				     adjlB.begin(); arcB!=adjlB.end(); ++arcB) {
				size_type ar = arcA->right();
				size_type br = arcB->right();
				
				if (
				    (params->max_diff < 0
				     || std::max(ar, br)-std::min(ar, br) <= (size_type)params->max_diff
				     )
				    &&
				    (params->max_diff_am < 0
				     || std::max(ar-al, br-bl)-std::min(ar-al, br-bl) <= (size_type)params->max_diff_am
				     )
				    )
				    // note: std::max(ar-al, br-bl)-std::min(ar-al, br-bl) is used
				    // since abs((ar-al) - (br-bl)) does not work for unsigned type
				    {
					max_ar=max(max_ar, ar);
					max_br=max(max_br, br);
				    }
			    }
			}
		    }

		    // ------------------------------------------------------------
		    // align under the maximal pair of arcs
		    //
		    align_in_arcmatch(al, max_ar, bl, max_br);

		    // ------------------------------------------------------------
		    // fill D matrix entries
		    //
		    fill_D_entries_inside(al, bl, adjlA, adjlB, max_ar, max_br);

		}

	    }
	}

    }

	D_created=true; // now the matrix D is built up
}

//===========================================================================
//	align_inside()
//===========================================================================	
//! compute the alignment score

pf_score_t AlignerP::align_inside() {
	// ------------------------------------------------------------
	// computes D matrix and then
	// does the alignment on the top level
	// ------------------------------------------------------------

	if (!D_created) {// std::cout<< "D is going to be created "<<endl; 
		align_D_inside();
	}

	assert(r.get_startA()>0);
	assert(r.get_startB()>0);

	align_in_arcmatch(r.get_startA()-1, r.get_endA()+1, r.get_startB()-1,r.get_endB()+1);
	
	partFunc = M(r.get_endA(), r.get_endB());
	
	assert(partFunc>0);
	
	return partFunc;
}

//===========================================================================
//	align_DPrime_outside()
//===========================================================================	
// compute all entries DPrime
void AlignerP::align_DPrime_outside() {
    // ------------------------------------------------------------
    // General workflow:
    //
    // for all combinations of left arc ends al and ar
    // 1.) determine for which arc-pairs the DPrime entries can be computed   
    // in one run, 2.) call align_out_arcmatch 3.) call fill_D_entries_outside
    // ------------------------------------------------------------

    // ------------------------------------------------------------
    // traverse the left ends al,bl of arcs in descending order
    //

    for (size_type al=r.get_startA(); al<=r.get_endA(); al++) {

	const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(al);

	if (adjlA.size()>0) {
	    
	    size_type min_bl=r.get_startB();
	    size_type max_bl=r.get_endB();

	    // if max-diff heuristic is active,
	    // restrict range for left ends of bl
	    if (params->max_diff>=0) {
		min_bl = std::max(min_bl+params->max_diff, al) - params->max_diff;
		max_bl = std::min(max_bl, al + params->max_diff);
	    }
	    
	    for (size_type bl=min_bl; bl<=max_bl; bl++) {
		const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(bl);
		if (adjlB.size()>0) {
		    // ------------------------------------------------------------
		    // get minimal right ends of arcs with left ends al,bl 
		    // where am-heuristic conditions hold
		    //
		    size_type min_ar=r.get_endA();
		    size_type min_br=r.get_endB();

		    if (am_heuristic) {
			min_ar=adjlA[0].right();
			min_br=adjlB[0].right();
		    } else {
			for (BasePairs::LeftAdjList::const_iterator arcA =
				 adjlA.begin(); arcA!=adjlA.end(); ++arcA) {
			    for (BasePairs::LeftAdjList::const_iterator arcB =
				     adjlB.begin(); arcB!=adjlB.end(); ++arcB) {

				size_type ar = arcA->right();
				size_type br = arcB->right();
				if (
				    (params->max_diff < 0
				     || std::max(ar, br)-std::min(ar, br) <= (size_type)params->max_diff
				     )
				    &&
				    (params->max_diff_am < 0
				     ||std::max(ar-al, br-bl)-std::min(ar-al, br-bl) <= (size_type)params->max_diff_am
				     )
				    )
				    {
					min_ar=min(min_ar, ar);
					min_br=min(min_br, br);
				    }
			    }
			}
		    }

		    // ------------------------------------------------------------
		    // align outside the arc
		    align_out_arcmatch(al, min_ar, bl, min_br);

		    // ------------------------------------------------------------
		    // fill DPrime matrix entries
		    //
		    fill_DPrime_entries_outside(al, bl, adjlA, adjlB, min_ar, min_br);

		}

	    }
	}

    }
    DPrime_created=true; // now the matrix DPrime is built up
}

//===========================================================================
//	init_MPrime()
//===========================================================================	
// generic initalization method. The last row and column are initialized.
// Initialize for the computation of outside alignments for all holes that start with
// left ends al,bl and end with right ends ar<=i<=lenA and br<=j<=lenB
void AlignerP::init_MPrime(size_type al, size_type ar, size_type bl, size_type br) {

    assert(al>0 && bl>0);

    MPrime(r.get_endA(),r.get_endB())=M(al-1,bl-1);
    //std::cout<<"in init_MPrime :  with al,ar,bl,br "<<al<<" "<<ar<< " : "<<bl<<" "<<br<<endl;
    
    for (int i=(int)r.get_endA()-1;i>=(int)ar; i--) {
	assert(MPrime(i+1,r.get_endB())>0);
	MPrime(i,r.get_endB()) = MPrime(i+1,r.get_endB()) * scoring->exp_gapA(i+1,r.get_endB());
    }
    
    for (int j=(int)r.get_endB()-1; j>=(int)br; j--) {
	assert(MPrime(r.get_endA(),j+1)>0);
	MPrime(r.get_endA(),j) = MPrime(r.get_endA(),j+1) * scoring->exp_gapB(r.get_endA(),j+1);;
    }
}



//===========================================================================
//	init_Mrev
//===========================================================================	
// init alignment matrix for aligning the fragments al..ar-1 and bl..br-1 
void AlignerP::init_Mrev(size_type al, size_type ar, size_type bl, size_type br) {
    assert(ar>0);
    assert(br>0);
    
    Mrev(ar,br)=(pf_score_t)1;
    
    for (size_type i=ar-1; i>=al; i--) {
	assert(Mrev(i+1,br)>0);
	Mrev(i,br) = Mrev(i+1,br) * scoring->exp_gapA(i,br);;
    }
    
    for (size_type j=br-1; j>=bl; j--) {
	assert(Mrev(ar,j+1)>0);
	Mrev(ar,j)= Mrev(ar,j+1) * scoring->exp_gapB(ar,j);
    }
}



//===========================================================================
//	align_noex_reversedM()
//===========================================================================	
// standard cases in alignment for reversed case: base match, base in/del, arc match
// compute pf of alignments i..al-1 and j..bl-1 
inline
pf_score_t
AlignerP::align_noex_reversedM(size_type i, size_type j,size_type al, size_type bl) {

	pf_score_t max_score;
	// base match
	//assert(Mrev(i+1,j+1)>0);

	max_score = Mrev(i+1,j+1) * scoring->exp_basematch(i,j);

	// base del
	assert(Mrev(i+1,j)>0);
	max_score += Mrev(i+1,j) * scoring->exp_gapA(i,j-1);

	// base ins
	assert(Mrev(i,j+1)>0);
	max_score += Mrev(i,j+1) * scoring->exp_gapB(i-1,j);

	// arc match
	// standard case for arc match (without restriction to lonely pairs)
	const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(i);
	const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(j);

	// for all pairs of arcs in A and B that have right ends i and j, respectively
	//
	for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin();
	arcA!=adjlA.end() && arcA->right() < al; ++arcA) {
		for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin();
		arcB!=adjlB.end() && arcB->right() < bl; ++arcB) {

			// consider score for match of basepairs
			assert(Mrev(arcA->right()+1,arcB->right()+1)>0);
			assert(am_heuristic || D(arcA->idx(),arcB->idx())>0);
			
			max_score +=
			    D(arcA->idx(),arcB->idx()) * Mrev(arcA->right()+1,arcB->right()+1);
		}
	}
	return max_score;
}


//===========================================================================
//	align_noex_outside()
//===========================================================================	
//! standard cases in alignment: base match, base in/del, arc match
//! compute partition function for alignments with hole al..i and bl..j (recursing to larger holes)
inline
pf_score_t 
AlignerP::align_noex_outside(size_type al, size_type bl, size_type i, size_type j) {

	pf_score_t max_score;

	// base match
	max_score = MPrime(i+1, j+1) * scoring->exp_basematch(i+1, j+1);

	// base del
	max_score += MPrime(i+1, j) * scoring->exp_gapA(i+1, j);

	// base ins
	max_score += MPrime(i, j+1) * scoring->exp_gapB(i, j+1);

	// arc match, case 4
	{
		const BasePairs::RightAdjList &adjlA = bpsA.right_adjlist(i+1);
		const BasePairs::RightAdjList &adjlB = bpsB.right_adjlist(j+1);

		// for all pairs of arcs in A and B that have right ends i+1 and j+1, respectively
		//

		for (BasePairs::RightAdjList::const_reverse_iterator arcA=adjlA.rbegin();
		arcA!=adjlA.rend() && arcA->left() < al; ++arcA) {
			for (BasePairs::RightAdjList::const_reverse_iterator arcB=adjlB.rbegin();
			arcB!=adjlB.rend() && arcB->left() < bl; ++arcB) {
				// consider score for match of basepair
				//std::cout<<"Mrev(arcA->left()+1,arcB->left()+1) "<<Mrev(arcA->left()+1,arcB->left()+1)<<endl;

				assert(am_heuristic || DPrime(arcA->idx(),arcB->idx()) > 0);
				// assert(Mrev(arcA->left()+1,arcB->left()+1) > 0);

				max_score += DPrime(arcA->idx(),arcB->idx()) * scoring->exp_arcmatch(*arcA,*arcB) * Mrev(arcA->left()+1,arcB->left()+1);

			}
		}
		//std::cout<<"Max score of outside up to case 4: " << max_score <<"  "<<al<<"  "<<bl<<"  "<<i<<"  "<<j<<endl;
	}

	// arc match, case 5
	{
		const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(i+1);
		const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(j+1);

		// for all pairs of arcs in A and B that have left ends i and j, respectively
		for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin(); arcA!=adjlA.end(); ++arcA) {
			for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin(); arcB!=adjlB.end(); ++arcB) {
				// consider score for match of basepairs

				assert(MPrime(arcA->right(),arcB->right())>0);
				assert(am_heuristic || D(arcA->idx(),arcB->idx())>0);
				
				max_score += MPrime(arcA->right(),arcB->right()) * D(arcA->idx(),arcB->idx());
			}
		}
	}

	//std::cout<<"Max score of outside up to case 5: " <<"  "<<al<<"  "<<bl<<"  "<<i<<"  "<<j<<endl;
	return max_score;
}





//===========================================================================
// align_out_arcmatch
//===========================================================================	
// fill MPrime(i,j) entries for fixed left indices al and bl
// where i>=ar and j>=br
void
AlignerP::align_out_arcmatch(size_type al,size_type ar,size_type bl,size_type br) {
	//std::cout<<"in align out arcmatch : calling init_out "<<endl;
    
    assert(al>0);
    assert(bl>0);
    
    init_MPrime(al,ar,bl,br);
    init_Mrev(r.get_startA(), al, r.get_startB(), bl);
    
    for(size_type i=al-1; i>=r.get_startA();i--) {
	for(size_type j=bl-1; j>=r.get_startB();j--) {
	    Mrev(i,j)=align_noex_reversedM(i,j,al,bl);
	}
    }
    
    for(int i=(int)r.get_endA()-1; i>=(int)ar; i--) {
	for(int j=(int)r.get_endB()-1;j>=(int)br; j--) {
	    MPrime(i,j) = align_noex_outside(al,bl,i,j);
	}
    }
}

//===========================================================================
//	fill_DPrime_entries_outside
//===========================================================================	

// compute the entries in the DPrime matrix that
// can be computed from the matrix/matrices MPrime
// for the subproblem al,bl,min_ar,min_br
//
// pre: MPrime matrices are computed by a call to 
//      align_out_arcmatch(al,min_ar,bl,min_br)
//
void
AlignerP::fill_DPrime_entries_outside(size_type al, size_type bl,
		const BasePairs::LeftAdjList &adjlA,
		const BasePairs::LeftAdjList &adjlB,
		size_type min_ar, size_type min_br)
{

	for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin();
	arcA!=adjlA.end(); ++arcA) {

		for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin();
		arcB!=adjlB.end(); ++arcB) {

			size_type ar = arcA->right();
			size_type br = arcB->right();
			//
			// if right ends ar,br exceed the limits min_ar,min_br resp.
			// the corresponding DPrime entry is set to zero
			// in order to dissalow the arc match completely.
			// This occurs only due to the am heuristic.
			if (ar<min_ar || br<min_br ) {
			    DPrime(arcA->idx(),arcB->idx()) = (pf_score_t)0;
			} else {
			    assert( MPrime(ar,br)>0 );
			    DPrime(arcA->idx(),arcB->idx()) = MPrime(ar,br);
			}
		}
	}
}

//===========================================================================
//	align_outside()
//===========================================================================	
// perform outside algo, compute DPrime
void AlignerP::align_outside() {
	// ------------------------------------------------------------
	// computes DPrime matrix and then
	// does the alignment on the top level
	// ------------------------------------------------------------

	if (!DPrime_created) { //std::cout<< "DPrime is going to be created "<<endl; 
		align_DPrime_outside();}
	
	//align_out_arcmatch(r.get_startA()-1,r.get_startA()-1,r.get_startB()-1,r.get_startB()-1);
}

//===========================================================================
// Align_rev
//===========================================================================
// compute alignment matrix for aligning the fragments al..ar-1 and bl..br-1
void AlignerP :: align_rev(size_type al, size_type bl,size_type ar, size_type br){
	assert(al<=ar);
	assert(bl<=br);

	init_Mrev(al,ar,bl,br);
	//std::cout<<"MRev after init call"<<endl;
	//									cout<<Mrev;
	//									std::cout<<endl;


	for(size_type p=ar-1; p>=al; p--) {
	    for(size_type q=br-1; q>=bl; q--) {
		Mrev(p,q)=align_noex_reversedM(p,q,ar,br);
	    }
	}
	
	//std::cout<<"align rev out returns for "<<Mrev(al,bl)<<" al,bl "<<al<<"   "<<bl<<Mrev(i-1,j-1)<<endl;
}

//===========================================================================
//compute_arcmatch_probabilities()
//===========================================================================
void 
AlignerP::compute_arcmatch_probabilities() {
	
    for(size_type i=0;i<bpsA.num_bps();i++){
	for(size_type j=0;j<bpsB.num_bps();j++){
	    
	    assert(am_heuristic || DPrime(i,j)>0);
	    assert(am_heuristic || D(i,j)>0);
	    assert(M(r.get_endA(),r.get_endB())>0);
	    
	    am_prob(i,j)=
		(D(i,j)/(long double)partFunc)
		*  DPrime(i,j);
	    
	    assert(am_prob(i,j)<=1);
	}
    }
    
    //std::cout<<"Arc match probs calculated"<<endl;
    //std::cout<<am_prob<<std::endl;
}

//===========================================================================
// compute_basematch_probabilities()
//===========================================================================
// pre: arc match probabilites am_prob are already computed
void
AlignerP::compute_basematch_probabilities( bool basematch_probs_include_arcmatch )
{
    //std::cout<<"Skip compute_basematch_probabilities"<<std::endl;
    //return;

    // sum the conditional partition functions for all distinct cases
    
        // consider only arc matchs with a probability of at more than am_prob_threshold
    double am_prob_threshold=sqrt(params->min_am_prob); // use something quite conservative as threshold, such that user can still control this 
	
    // --------------------------------------------------
    // cases, where edge is enclosed by arc match
    //
    
    for(size_type al=r.get_startA();al<=r.get_endA();al++){
	for(size_type bl=r.get_startB();bl<=r.get_endB();bl++){
	    const BasePairs::LeftAdjList &adjlA = bpsA.left_adjlist(al);
	    const BasePairs::LeftAdjList &adjlB = bpsB.left_adjlist(bl);
	    //	if(al==r.get_endA() || bl==r.get_endB())
	    //							std::cout<<" ends encountered here"<<endl;
	    if(adjlA.size() >0 && adjlB.size()>0)
		{
			    
		    // get max_ar and max_br (due to am_prob)
		    
		    size_type max_ar=al;
		    size_type max_br=bl;
		    		    
		    for (BasePairs::LeftAdjList::const_iterator arcA = adjlA.begin();
			 arcA!=adjlA.end(); ++arcA) {
			for (BasePairs::LeftAdjList::const_iterator arcB = adjlB.begin();
			     arcB!=adjlB.end(); ++arcB)
			    {
				size_type ar = arcA->right();
				size_type br = arcB->right();
				
				if ( am_prob(arcA->idx(),arcB->idx()) > am_prob_threshold ) {
				    max_ar=max(max_ar, ar);
				    max_br=max(max_br, br);
				}
			    }
		    }
		    
		    
		    assert(D_created);assert(DPrime_created);
		    align_in_arcmatch(al,max_ar,bl,max_br);
		    
		    for (BasePairs::LeftAdjList::const_iterator arcA=adjlA.begin();
			 arcA!=adjlA.end(); ++arcA) {
			for (BasePairs::LeftAdjList::const_iterator arcB=adjlB.begin();
			     arcB!=adjlB.end(); ++arcB) {
			    
			    if (am_prob(arcA->idx(),arcB->idx()) > am_prob_threshold) {
				
				size_type ar=arcA->right();
				size_type br=arcB->right();
				
				align_rev(al+1,bl+1,ar,br);
				
				// a part of the pf-contrib can be computed outside of the loops
				pf_score_t arcmatch_outside_pf=
				    scoring->exp_arcmatch(*arcA,*arcB)
				    * DPrime(arcA->idx(),arcB->idx());
				
				// add contributions for all alignment edges enclosed by the arc match (arcA,arcB)
				for(size_type i=al+1;i<ar;i++){
				    for(size_type j=bl+1;j<br;j++){
					
					//	std::cout<<" calling rev outside for "<<i<<"  "<<j<<" having "<<ar<<"  "<<br<<endl;
					
					assert( M(i-1,j-1)>0 );
					assert( Mrev(i+1,j+1)>0 );
					
					bm_prob (i,j) += 
					    M(i-1,j-1)
					    * scoring->exp_basematch(i,j)
					    * Mrev(i+1,j+1)
					    * arcmatch_outside_pf;
				    }
				}
			    }
			}
		    }
		}
	    //	else{std::cout<<"left adj list empty " <<al<<"  "<<bl<<endl;}
	}
    }

    // --------------------------------------------------
    // extra case, where there is no enclosing arc match of alignment edge (i,j)
	
    align_in_arcmatch(0,r.get_endA()+1,0,r.get_endB()+1);
    align_rev(1,1,r.get_endA()+1,r.get_endB()+1);
	
    for(size_type i=1;i<=r.get_endA();i++){
	for(size_type j=1;j<=r.get_endB();j++){
	    
	    assert(M(i-1,j-1)>0);
	    assert(Mrev(i+1,j+1)>0);
	    
	    bm_prob(i,j) += M(i-1,j-1) * scoring->exp_basematch(i,j) * Mrev(i+1,j+1);
	}
    }
	
    //std::cout<<"Matrix basematch probabilities "<<endl;
	
    // --------------------------------------------------
	
    if ( basematch_probs_include_arcmatch ) {
	    
	for(size_type i=0;i<bpsA.num_bps();i++){
	    for(size_type j=0;j<bpsB.num_bps();j++){
		    
		assert(am_heuristic || DPrime(i,j)>0);
		assert(am_heuristic || D(i,j)>0);
		
		pf_score_t pf = D(i,j) *  DPrime(i,j);
		
		bm_prob(bpsA.arc(i).left(),bpsB.arc(j).left()) += pf;
		bm_prob(bpsA.arc(i).right(),bpsB.arc(j).right()) += pf;
	    }
	}
    }

    // --------------------------------------------------
    // divide the conditional partition functions by total partition function  
	
	for(size_type i=1;i<=r.get_endA();i++){
		for(size_type j=1;j<=r.get_endB();j++){
			bm_prob(i,j)=bm_prob(i,j)/partFunc;
			assert(bm_prob(i,j)<=1);
		}
	}
}


//===========================================================================
//write_basematch_probabilities()
//===========================================================================
void AlignerP :: write_basematch_probabilities(std::ostream &out)
{
    for(size_type i=1;i<=r.get_endA();i++){
	for(size_type j=1;j<=r.get_endB();j++){
	    if (bm_prob(i,j)>=params->min_bm_prob) {
		out<<i<<" "<<j<<" "<<bm_prob(i,j);
		out<<std::endl;
	    }
	}
    }
}

//===========================================================================
//write_arcmatch_probabilities()
//===========================================================================
void AlignerP :: write_arcmatch_probabilities(std::ostream &out)
{
	for(size_type i=0; i<bpsA.num_bps();i++){
		for(size_type j=0; j<bpsB.num_bps();j++){
		    if (am_prob(i,j)>=params->min_am_prob) {
			out<<bpsA.arc(i).left()<<" "<<bpsA.arc(i).right()<<" "
			   <<bpsB.arc(j).left()<<" "<<bpsB.arc(j).right()<<" "
			   <<am_prob(i,j);
			out<<std::endl;
		    }
		}
	}
}
